#  CentroidErrorPlot23.py
#  Plots tabular data of SigmaX vs Rmag from SIMv23.py
#  36 cases: 9 Rmags, two background noise levels, two seeings
#  100 noise iterations for each case.
#  Filename is SIM23-36cases.dat
"""
0 given BkgRMS
1 given seeing
2 given Rmag
3 given Peak
4 fit Bkg
5 fit Peak
6 fit Xbar
7 fit Ybar
8 fit FWHM
9 fit Total
10 fit BkgRMS
11 fit PeakRMS
12 fit XbarRMS
13 fit YbarRMS
14 fit FWHM-RMS
15 fit TotalRMS
16 derived PeakSNR
17 derived TotalSNR
18 derived PeakMonetCoef
19 derived TotalMonetCoef
"""
# Lab data are from....
# ref:  R. Casas et al, 12 Oct 2018, 'Analysis of BIAS, CTE, DARK, and PTC' for GFA#05
#  this uses CCD e2v 13072-24-01
# GFA5 raw FITS images on NERSC:
#   /global/projecta/projectdirs/desi/engineering/gfa/PreProductionCamera/Data/2018101_GFA5
#   Each image is 37MB = four 1MPIX quadrants * 8 bytes/pixel
#   The FATandFAINT star images are on quadrant 3. Two bias frames, then ten 1-sec exposures.
# Reports on NERSC:
#   /global/projecta/projectdirs/desi/engineering/gfa/PreProductionCamera/Data/99999999_GFA_FINAL
#   page 4 lists reciprocal gains; for Amp#3 is is 3.644 e/ADU hence G=0.274 ADU/electron
#
import numpy as np
import matplotlib.pyplot as plt  
import matplotlib.patches as patches

PROGNAME = 'CentroidErrorPlot23.py'
filename = 'SIM23-36cases.dat'  # from EXCEL and .csv
SOURCENAME = 'from ' + filename
SYSTEMSPEC = 0.14  # rms pixels which is 0.03 arcseconds RMS

results = np.loadtxt(filename, skiprows=2)

print()

xvals6g  = results[:9,2]   # Rmag list, top 9 elements, good seeing
yvals6g  = results[:9,13]  # Yrms list, top 9 elements, good seeing

xvals16g  = results[9:18,2]   # Rmag list, 9 to 18, good seeing
yvals16g  = results[9:18,13]  # Yrms list, 9 to 18, good seeing

xvals6b = results[18:27,2]   # Rmag list, 18 to 27, bad seeing
yvals6b = results[18:27,13]  # Yrms list, 18 to 27, elements

xvals16b = results[27:,2]   # Rmag list, bottom 9 elements, bad seeing
yvals16b = results[27:,13]  # Yrms list, bottom 9 elements, bad seeing


fig, ax = plt.subplots()

plt.semilogy(xvals16b, yvals16b, '-ko')  # worst at top
plt.semilogy(xvals6b, yvals6b, '-go')    # slightly better
plt.semilogy(xvals16g, yvals16g, '-ro')  # better still
plt.semilogy(xvals6g, yvals6g, '-bo')    # best at bottom
plt.semilogy((15.0,17.0),(0.15,0.15))
ax.text(15.0, 0.155, 'System Spec')


ax.set_ylim(ymin=0.001, ymax=0.5)
ax.set_xlabel('Rband magnitude')
ax.set_ylabel('RMS Centroid Error, pixels')
ax.tick_params(which='both', direction='in')  # Yay, works even for log minor ticks.
ax.text(0.05, 0.92, 'BkgRMS=16ADU, 1.6arcsec', transform=ax.transAxes, fontsize=10, color='black')
ax.text(0.05, 0.88, 'BkgRMS=6ADU, 1.6arcsec', transform=ax.transAxes, fontsize=10, color='green')
ax.text(0.05, 0.84, 'BkgRMS=16ADU, 0.8arcsec', transform=ax.transAxes, fontsize=10, color='red')
ax.text(0.05, 0.80, 'BkgRMS=6ADU, 0.8arcsec', transform=ax.transAxes, fontsize=10, color='blue')
ax.text(0.65, 0.06, PROGNAME, transform=ax.transAxes, fontsize=8)
ax.text(0.65, 0.03, SOURCENAME, transform=ax.transAxes, fontsize=8)

targetRect = patches.Rectangle((13.9,0.01), 0.2, 0.005, edgecolor='red', facecolor='none', linewidth=4)
plt.annotate('Lab Data, 0.8arcsec', xy=(14.1,0.010), xytext=(14.4,0.005), arrowprops=dict(ec='red', fc='red'))
ax.add_patch(targetRect)

plt.title('Simulated Centroid Errors vs R-band Magnitude')

plt.savefig('CentroidErrorPlot23.png') 
plt.show()           


"""
fit17 = np.polyfit(xvals17, yvar17, deg=1) # this gets the best fit coefs
slope17Str = 'Rn=17e: slope,ADU/electron = {:5.3f}'.format(fit17[0])
fit17Func = np.poly1d(fit17)  # this is the callable polynomial

fit58 = np.polyfit(xvals58, yvar58, deg=1) # this gets the best fit coefs
slope58Str = 'Rn=58e: slope,ADU/electron = {:5.3f}'.format(fit58[0])
fit58Func = np.poly1d(fit58)  # this is the callable polynomial

plt.plot(xvals58, yvar58, 'ro', xvals58, fit58Func(xvals58), '-k')
"""
