# GFAv17.py   ...RUN TWICE: BIASSUB=TRUE and BIASSUB=FALSE
#             ...NEED BOTH TO USE THE TWO FnF data files
#             ...Crazy data limiter lines 540-570 found to be unnecessary
#
#  Like GFAv16 but exploring subpixel resolution within the Gaussian peak.
#  Motive: single pixel sample leads to overestimate on convex surfaces
#     and underestimate on concave surfaces.  For a Gaussian, the central part
#     is overestimated and the skirts are underestimated, possibly messing up the
#     derived Xpeak, Ypeak, Zpeak, and FWHM.
#
# Reads, displays, and fits a 2D Gaussian to a 40x40 thumbnail extracted from a FITS file.
# M.Lampton UCB SSL Oct 2018 v13, simpified for FAT-and-FAINT local file names
#
# ref:  R. Casas et al, 12 Oct 2018, 'Analysis of BIAS, CTE, DARK, and PTC' for GFA#05
#  this uses CCD e2v 13072-24-01
# GFA5 raw FITS images on NERSC:
#   /global/projecta/projectdirs/desi/engineering/gfa/PreProductionCamera/Data/2018101_GFA5
#   Each image is 37MB = four 1MPIX quadrants * 8 bytes/pixel
#   The FATandFAINT star images are on quadrant 3. Two bias frames, then ten 1-sec exposures.
# Reports on NERSC:
#   /global/projecta/projectdirs/desi/engineering/gfa/PreProductionCamera/Data/99999999_GFA_FINAL
#   page 4 lists reciprocal gains; for Amp#3 is is 3.644 e/ADU hence G=0.274 ADU/electron
#
# Now including the integrated Gaussian "total" diagnostic.
# Now with or without bias frame subtraction
#
# For use with SNRv16.py, RUN THIS TWICE: with, then without, bias subtraction
#
# Now without trimming; EDGE=0.

# DANGER THERE IS A SINGLE HOT PIXEL in IMAGE=008, 009; 4347 ADU tall. 
#   at untrimmed {687,530} or trimmed={666,509} or thumbnail offsets={646,489}.
# Best to predefine a single fixed thumbnail zone for analysis
#   use trimmed offsets={776,804} to capture the star peak at trimmed={796,824}

# for GFAv16.py:
# To do: raise the text display within the thumbnail map graphics.   << DONE
# to do: extract and evaluate the longitudinal pixel noise from the aggregate 10 images <<DONE
# To do: format output table for LaTeX    << DONE
#
# Casas, R., Oct 12th 2018, on NERSC
# "This document is a summary of the analysis of the BIAS, DARK, PTC and CTE
# images obtained in October 11th, 2018 of the GFA#05 cycled unit equipped with
# the CCD e2v 13072-24-01 and the stepped filter id#2.
# The images used to do these test can be found in the NERSC directory:
# /global/projecta/projectdirs/desi/engineering/gfa/PreProductionCamera/
# Data/2018101_GFA5
#
# Night sky and R-band magnitude data are from "GuiderSNR-v20.xlsx" Oct 2018
# RbandMag = 15.0 -2.5*log(e/sec / 53400)
# e/Sec = 53400 * 10^(0.4*(15-RbandMag))
# Note: Casas et al find GAIN= 0.274 ADU/electron
# Note: GFAv16.py finds total ADU volume = 36540 ADU = 133361 electrons, 1 sec
# This is RbandMag= 14.006


import sys
from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt  
from scipy.optimize import curve_fit

PROGNAME = 'GFAv17'

BIASSUBTRACT = False
# BIASSUBTRACT = True

THUMBSIDE = 40                #  square thumbnail edge size, pixels 
THUMBPIXELS = THUMBSIDE**2    # how many pixels in a thumbnail

THUMBMINCOL = 796  # Xoffset in raw image
THUMBMAXCOL = THUMBMINCOL+THUMBSIDE
THUMBMINROW = 824  # Yoffset in raw image
THUMBMAXROW = THUMBMINROW+THUMBSIDE

NPARMS = 5    # for fitting a 2D Gaussian
headerstr = '     Bkg/pix  PeakAboveBkg   Xpeak       Ypeak         FWHM       Total       BkgAve        BkgRMS'


#-------declare the image model for fitting----------------------------

def myGFA16Func2D(parms, nrows, ncols):
    # GFAv16 uses just one sample point per pixel.
    # Produces a noiseless image, nrows x ncols, from given parameters.
    # parms[]: {background/pix;  peakSignal/pix; xpeak; ypeak; fwhmpix}
    result = np.zeros((nrows,ncols))
    for row in range(nrows):
        for col in range(ncols):
            r2 = (col-parms[2])**2 + (row-parms[3])**2
            arg = 0.5*r2/(parms[4]/2.35482)**2
            result[row,col] = parms[0] + parms[1]*np.exp(-arg)
    return result
    
    
def myHelperFunc2D(parms, nrows, ncols):
    # Produces a noiseless image, nrows x ncols, from given parameters.
    # parms[]: {background/pix;  peakSignal/pix; xpeak; ypeak; fwhmpix}
    denom = 2.*(parms[4]/2.35482)**2    
    result = np.zeros((nrows,ncols))
    result.fill(parms[0])
    for row in range(nrows):
        for col in range(ncols):
            r2 = (col-parms[2])**2 + (row-parms[3])**2
            arg = -r2/denom
            if arg > -10:                                 # TEST FOR SIGNIFICANCE
                result[row,col] += parms[1]*np.exp(arg)   # NOTICE THE "+="
    return result
    

    
NSUBS = 25
devs = np.array([[-0.4, -0.4],
                 [-0.4, -0.2],
                 [-0.4,  0.0],
                 [-0.4,  0.2],
                 [-0.4,  0.4],
                 [-0.2, -0.4],
                 [-0.2, -0.2],
                 [-0.2,  0.0],
                 [-0.2,  0.2],
                 [-0.2,  0.4],
                 [ 0.0, -0.4],
                 [ 0.0, -0.2],
                 [ 0.0,  0.0],
                 [ 0.0,  0.2],
                 [ 0.0,  0.4],
                 [ 0.2, -0.4],
                 [ 0.2, -0.2],
                 [ 0.2,  0.0],
                 [ 0.2,  0.2],
                 [ 0.2,  0.4],
                 [ 0.4, -0.4],
                 [ 0.4, -0.2],
                 [ 0.4,  0.0],
                 [ 0.4,  0.2],
                 [ 0.4,  0.4]])
    
def myFunc2D(parms, nrows, ncols):
    # Uses 25 samples per pixel, using devs[] from center {col,row}.
    # Produces a noiseless image, nrows x ncols, from given parameters.
    # parms[]: {background/pix;  peakSignal/pix; xpeak; ypeak; fwhmpix}
    denom = 2.*(parms[4]/2.35482)**2
    result = np.zeros((nrows,ncols))
    result.fill(parms[0])
    for row in range(nrows):
        for col in range(ncols):
            r2 = (col-parms[2])**2 + (row-parms[3])**2
            arg = -r2/denom
            if arg > -10.0:
                subvalues = np.zeros(NSUBS)
                for sub in range(NSUBS):
                    r2 = (col + devs[sub,0] - parms[2])**2 + (row + devs[sub,1] - parms[3])**2
                    arg = -r2/denom
                    subvalues[sub] = np.exp(arg)
                result[row,col] += parms[1]*np.mean(subvalues)  # NOTICE THE "+="
    return result
    

#---Optimization helpers: user should call this minimize2D() not curve_fit()-----
    
    

def minimize2D(goals2D, pinit1D, sigmas2D):
    # This is called by user to minimize.  
    #  It relies on wrapfunc1D() to always call myFunc2D().
    # Converts user input 2D arrays into 1D for calling curve_fit().
    # Also provides required args1D[] but we ignore them and use pixel {row,col}
    args1D = np.arange(THUMBPIXELS) 
    goals1D = goals2D.reshape(THUMBPIXELS)
    sigmas1D = sigmas2D.reshape(THUMBPIXELS)
    popt, pcov = curve_fit(wrapfunc1D, args1D, goals1D, pinit1D, sigmas1D)
    return popt, pcov
    
def wrapfunc1D(args, *parms): 
    # Removing the asterisk below gives us our parameter array.
    # args[] is discarded since we use pixel {row,col} as arguments.  
    # global GNROWS, GNCOLS must be preset by minimize2D().
    global gncalls
    gncalls += 1
    model2D = myFunc2D(parms, THUMBSIDE, THUMBSIDE)
    model1D = np.reshape(model2D, THUMBPIXELS)
    return model1D



def getRawImage(filename, quadrant):
    hdulist = list()
    try:
        hdulist = fits.open(filename)
    except IOError:
        print('Quitting, since failed to open ' + filename )
        quit()

    print('Addressing quadrant = ', quadrant)
    hdu = hdulist[quadrant]
    rawimage = hdu.data
    
    print('Properties of the FITS raw image in this quadrant:')
    NROWS = len(rawimage)
    NCOLS = len(rawimage[0])
    NPIXELS = NCOLS * NROWS    
    print('NCOLS, NDOWS:   ', NROWS, NCOLS)
    

    """
    #-------Plot the entire raw image---------------
    fig = plt.figure(figsize=(6,6))   # inches
    image = plt.imshow(trimmed)
    title = filename + '_quad_' + str(quadrant)
    plt.title(filename)
    plt.savefig(title + '.png')
    plt.show()
    """

    print('Sum ADU:     {:12.0f}'.format(np.sum(rawimage)))
    print('Average ADU: {:12.3f}'.format(np.mean(rawimage)))
    print('Median ADU:  {:12.3f}'.format(np.median(rawimage)))
    print('Max ADU:     {:12.3f}'.format(np.max(rawimage)))
    print('Min ADU:     {:12.3f}'.format(np.min(rawimage)))
    return rawimage


def build1200():   # full 1200 pixel bkg
    pixList = list()    
    for ipix in range(1600): 
        row, col = np.unravel_index(ipix, (40, 40)) 
        if row<10 or row>=30 or col<10 or col>=30:
            pixList.append(ipix)
    print('build1200() found N=', len(pixList))
    return pixList


def build700():       # outermost 700 pixels of the 1200
    # construct a list of 1D pixel addresses into 40x40 image
    pixList = list() 
    for ipix in range(1600):  
        row, col = np.unravel_index(ipix, (40, 40))
        if row<5 or row>=35 or col<5 or col>=35:
            pixList.append(ipix)
    print('build700() found N=', len(pixList))
    return pixList


def build500():    # innermost 500 pixels of the 1200
    # construct a list of 1D pixel addresses into 40x40 image
    pixList = list()    
    for ipix in range(1600): 
        row, col = np.unravel_index(ipix, (40, 40))
        if row>=5 and row<35 and col>=5 and col<35:
            if row<10 or row>=30 or col<10 or col>=30: 
                pixList.append(ipix)
    print('build500() found N=', len(pixList))
    return pixList




def analyze(filename, quadrant):
    # Extracts rawimage[] from quadrant subfile
    # biasAve[][] is supplied globally by host
    # Creates working[] from rawimage[]
    # Extracts thumb[] from working[]
    # Analyzes thumb[]
    # Extracts local background from thumb[]
    # Gets mean & variance of local background
    # returns best fit results and thumb[] itself
    global gncalls   # for iterative fitting

    if BIASSUBTRACT: 
        working = getRawImage(filename, quadrant) - biasAve
    else:
        working = getRawImage(filename, quadrant)
    
    RAWNROWS = len(working)
    RAWNCOLS = len(working[0])
    RAWNPIXELS = RAWNCOLS * RAWNROWS
    
    #-----extract the thumbnail----------

    thumb = working[THUMBMINROW:THUMBMAXROW, THUMBMINCOL:THUMBMAXCOL]
    print('thumb.shape = ', thumb.shape)   # should be SIDE x SIDE
    shortname = filename + '_quad' + str(quadrant)
    print(shortname)
    
    #-------Let's fit one 5-parm Gaussian to the thumbnail peak----------------

    peakrow, peakcol = np.unravel_index(thumb.argmax(), thumb.shape)
    averageADU = np.average(thumb)
    medianADU  = np.median(thumb)
    maxADU     = np.max(thumb)
    noise      = np.std(thumb)
    height     = maxADU - averageADU
    back       = averageADU
    print('...Starting Gaussian fit to the thumbnail peak:')
    gncalls = 0  
    myPinit    = np.array([back, height, peakcol, peakrow, 5.])  # an initial guess
    mySigmas2D = noise*np.ones([THUMBSIDE,THUMBSIDE])            # array of expected errors
    myData2D   = thumb                                           # the noisy image 
    optim, covar = minimize2D(myData2D, myPinit, mySigmas2D)     # uses global THUMBSIDE
    print('ncalls = ', gncalls)
    totalFit = 1.1331 * optim[1] * optim[4]**2                   # integral of the best fit Gaussian   
     
    #----extract and characterize local 1200 pixel background----------------
    #--------yikes it has extreme pixels: min=162ADU, max=280ADU------------
    #-----let's apply some kind of filter?------------------------------
    
    bkgPixList = build1200()
    bkglen = len(bkgPixList)  
    print('bkglen = ', bkglen)  # should be 1200
    
    thumb1D = np.reshape(thumb, THUMBPIXELS)
    bkg = np.zeros((bkglen))
    for pix in range(bkglen):
        bkg[pix] = thumb1D[bkgPixList[pix]]
    avebkg = np.average(bkg)
    stdbkg = np.std(bkg)
    minbkg = np.min(bkg)
    maxbkg = np.max(bkg)
    
    print('avebkg = {:12.3f}'.format(avebkg))
    print('stdbkg = {:12.3f}'.format(stdbkg))
    print('minbkg = {:12.3f}'.format(minbkg))
    print('maxbkg = {:12.3f}'.format(maxbkg))
    
    
    #----Now build the eight parms array summarizing this image---------

    print('Best Fit Parms to images in ' + shortname)
    eightParms = np.array([optim[0], optim[1], optim[2]+THUMBMINCOL, optim[3]+THUMBMINROW, optim[4], totalFit, avebkg, stdbkg])

    #----plot this extracted thumbnail showing offsets-----------------
    
    numberstr = ''
    for value in eightParms:
        numberstr += '{:12.3f}'.format(value)
        
    """
    fig = plt.figure(figsize=(6,6))
    ax = fig.add_subplot(111)
    image = plt.imshow(thumb)
    plt.colorbar(image, fraction=0.046, pad=0.04) # these numbers always "magically" work
    title = filename + '_quad=' + str(quadrant)
    plt.title(title)
    ax.set_xlabel('X - raw Xoffset = ' + str(THUMBMINCOL))
    ax.set_ylabel('Y - raw Yoffset = ' + str(THUMBMINROW))    
    plt.subplots_adjust(bottom=0.12)              # room for best fit parameters
    fig.text(0.04, 0.04, headerstr, fontsize=8)   # best fit parameters header
    fig.text(0.04, 0.02, numberstr, fontsize=8)   # best fit parameter values
    thumbfilename = title + '_' + PROGNAME + '.png'
    print('...saving thumbnail as ' + thumbfilename)
    plt.savefig(thumbfilename)
    plt.show()
    """
    
    #---summarize this images results---------
    outputstr = shortname + numberstr
    print(headerstr) 
    print(outputstr)    
    thumb1D = thumb.ravel()                       # flattens to 1D
    # thumb1D = np.reshape(thumb, THUMBPIXELS)    # flattens to 1D
    return eightParms, thumb1D                    # thumb1D 1600 pixels for vertical analysis
    





#--------------------------main program---------------------
gncalls = 0

biasnames = [
             'IMAGE_000.fits',
             'IMAGE_001.fits']             
                 
starnames = [
             'IMAGE_002.fits',
             'IMAGE_003.fits',
             'IMAGE_004.fits',
             'IMAGE_005.fits',
             'IMAGE_006.fits',
             'IMAGE_007.fits',
             'IMAGE_008.fits',  # hot pixel
             'IMAGE_009.fits',  # hot pixel
             'IMAGE_010.fits',
             'IMAGE_011.fits']

NSTARS = len(starnames)            

quadrant = 3

bias000 = getRawImage(biasnames[0], quadrant)
bias001 = getRawImage(biasnames[1], quadrant)
biasAve = 0.5*(bias000 + bias001)  # globally available to analyze()

if BIASSUBTRACT:
    outfilename = 'FnF_BiasSubtracted_' + PROGNAME + '.dat'
else:
    outfilename = 'FnF_NotBiasSubtracted_' + PROGNAME + '.dat' 
    
    
#------------ANALYZE THE IMAGES----------------------------

# prepare an array of thumbs1D[] for further analysis
# prepare an array to receive each fit numerical parameters

allThumbs = np.zeros((NSTARS, THUMBPIXELS))
allParms = np.zeros((NSTARS, 8))  # summary rows are added later into text form

for i in range(NSTARS):
    allParms[i], allThumbs[i] = analyze(starnames[i], quadrant)

#-----evaluate the vertical parm statistics----

aveRow = np.zeros(8)
rmsRow = np.zeros(8)
for whichparm in range(8):  # currently using 8 parms
    parms = allParms[:, whichparm]
    aveRow[whichparm] = np.average(parms)
    rmsRow[whichparm] = np.std(parms)
        
#--------------------now build the string table-----------------------
#---this edition uses '&' as field separators for LaTeX typesetting---
#----this edition uses double backslash for EOL, written '\\\\'-------

allParamStrings = list()
for i in range(NSTARS):
    myString = format(i+2, '03')
    for value in allParms[i]:
        myString += ' &{:12.3f}'.format(value)
    myString += ' \\\\'
    allParamStrings.append(myString)
    
myString = 'Col Aves'
for value in aveRow:
    myString += ' &{:12.3f}'.format(value)
myString += ' \\\\'
allParamStrings.append(myString)
    
myString = 'Col STDs'
for value in rmsRow:
    myString += ' &{:12.3f}'.format(value)
myString += ' \\\\'
allParamStrings.append(myString)

# should now have 12 rows: 10 stars, column averages, column RMSs
    
 
print()
print()
print('Summarizing ' + PROGNAME + ' with BIASSUBTRACT = ', BIASSUBTRACT)
print('for NSTARS   = ', NSTARS) 
for row in allParamStrings:
    print(row)

#-----Evaluate the statistics for each of the 1200 background pixel stacks----------

listOfStackAVEs = list()
listOfStackRMSs = list()                
listOfStackVARs = list()  
bkglist = list()

for ipix in range(THUMBPIXELS):                  # look at all 1600 thumbnail pixels
    row, col = np.unravel_index(ipix, (THUMBSIDE, THUMBSIDE))
    if row<10 or row>=30 or col<10 or col>=30:   # choose only surrounding background
        stack = allThumbs[:,ipix]                # stack of bkg pixel values
        for item in stack:
            bkglist.append(item)
        stackAVE = np.average(stack)             # average of each stack
        stackRMS = np.std(stack)                 # rms of each stack.  But: trends over time?
        stackVAR = np.var(stack)
        listOfStackAVEs.append(stackAVE)
        listOfStackRMSs.append(stackRMS)
        listOfStackVARs.append(stackVAR)

# print()
# print('len(LocalBKG-listOfAVEs) is ', len(listOfStackAVEs))
# print('len(LocalBKG-listOfRMSs) is ', len(listOfStackRMSs))  # YAY, 1200 numbers
# print('len(bkglist) :', len(bkglist))          # Yay, 12000 numbers

arrayOfStackAVEs = np.array(listOfStackAVEs)   # convert to an array for mean & var
arrayOfStackRMSs = np.array(listOfStackRMSs)   # convert to an array for mean & var
arrayOfStackVARs = np.array(listOfStackVARs)   # ditto

aveStackAVE = np.average(arrayOfStackAVEs)     # average of the 1200 stack averages
rmsStackAVE = np.std(arrayOfStackAVEs)         # rms of the 1200 stack averages; but gradients?
varStackAVE = np.var(arrayOfStackAVEs)

aveStackRMS = np.average(arrayOfStackRMSs)     # average of the 1200 stack rms's; but trends?
rmsStackRMS = np.std(arrayOfStackRMSs)         # rms of the 1200 stack rms's; but gradients?
varStackRMS = np.var(arrayOfStackRMSs)

aveStackVAR = np.average(arrayOfStackVARs)
rmsStackVAR = np.std(arrayOfStackVARs)
varStackVAR = np.var(arrayOfStackVARs)


ASA = '   aveStackAVE={:10.3f}'.format(aveStackAVE)
RSA = '   rmsStackAVE={:10.3f}'.format(rmsStackAVE)
VSA = '   varStackAVE={:10.3f}'.format(varStackAVE)

ASR = '   aveStackRMS={:10.3f}'.format(aveStackRMS)
RSR = '   rmsStackRMS={:10.3f}'.format(rmsStackRMS)
VSR = '   varStackRMS={:10.3f}'.format(varStackRMS)

ASV = '   aveStackVAR={:10.3f}'.format(aveStackVAR)
RSV = '   rmsStackVAR={:10.3f}'.format(rmsStackVAR)
VSV = '   varStackVAR={:10.3f}'.format(varStackVAR)

print()
print(ASA)
print(RSA)
print(VSA)
print()
print(ASR)
print(RSR)
print(VSR)
print()
print(ASV)
print(RSV)
print(VSV)
print()

with open(outfilename,'w') as outfile:
    outfile.writelines(ASA+RSA+VSA + ASR+RSR+VSR + ASV+RSV+VSV + '\n') 
    outfile.writelines(PROGNAME + '       ' + headerstr + '\n')
    for myString in allParamStrings:
        outfile.writelines(myString + '\n')

     
#--------Histogram "bkglist" of 12000 samples of background-------------

bkgarray = np.array(bkglist)
avebkg = np.mean(bkgarray)
rmsbkg = np.std(bkgarray)
avestr = 'ave = {:7.3f}'.format(avebkg)
rmsstr = 'rms = {:7.3f}'.format(rmsbkg)
print(avestr)   # finding 224.42, exactly as expected
print(rmsstr)   # finding 19.125, much higher than expected


if BIASSUBTRACT:
    note = 'with bias subtraction'
    histofilename = 'HistoBkg-WithBiasSub.png'
    lowerbin = 150
    upperbin = 300
    bins = range(lowerbin, upperbin, 1)
    # meanbin = 0.5*(lowerbin+upperbin)
    # for pix in bkglist:
    #     if pix<lowerbin or pix>upperbin:
    #        pix = mean
else:
    note = 'without bias subtraction'
    histofilename = 'HistoBkg-NoBiasSub.png'    
    lowerbin = 2075
    upperbin = 2225
    bins = range(lowerbin, upperbin, 1)
    # mean = 0.5*(lowerbin+upperbin)
    # for pix in bkglist:
    #     if pix<lowerbin or pix>upperbin:
    #        pix = mean
    
fig, ax = plt.subplots()    
plt.hist(bkglist, bins)
plt.title('Histogram of 12000 Measured Background Pixels')

ax.text(0.05, 0.92, note,  transform=ax.transAxes, fontsize=12)
ax.text(0.05, 0.87, avestr, transform=ax.transAxes, fontsize=12)
ax.text(0.05, 0.82, rmsstr, transform=ax.transAxes, fontsize=12)

ax.text(0.05, 0.05, PROGNAME+'.py', transform=ax.transAxes, fontsize=8)
ax.set_xlabel('ADU bin')
ax.set_ylabel('Pixels per ADU bin')

plt.savefig(histofilename) 
plt.show()


"""

Results from GFAv17 with 25 pixel subsamples:

Summarizing GFAv17 with BIASSUBTRACT =  True
for NSTARS   =  10
002 &     226.319 &    2421.338 &     816.576 &     844.113 &       3.639 &   36333.840 &     222.458 &      16.142 \\
003 &     220.627 &    2374.759 &     816.580 &     844.100 &       3.673 &   36307.734 &     217.039 &      16.697 \\
004 &     225.812 &    2383.223 &     816.567 &     844.119 &       3.682 &   36602.364 &     222.268 &      16.584 \\
005 &     228.907 &    2392.124 &     816.566 &     844.115 &       3.659 &   36285.915 &     225.303 &      16.808 \\
006 &     234.707 &    2369.898 &     816.543 &     844.079 &       3.684 &   36445.522 &     230.872 &      16.764 \\
007 &     231.458 &    2332.535 &     816.541 &     844.073 &       3.730 &   36770.214 &     228.122 &      16.961 \\
008 &     243.319 &    2219.559 &     816.494 &     844.023 &       3.829 &   36880.765 &     239.713 &      17.144 \\
009 &     239.735 &    2212.946 &     816.479 &     844.003 &       3.828 &   36746.992 &     236.029 &      16.757 \\
010 &     212.732 &    2248.784 &     816.471 &     844.004 &       3.780 &   36399.720 &     208.822 &      17.122 \\
011 &     216.920 &    2229.920 &     816.465 &     844.005 &       3.806 &   36601.908 &     213.593 &      16.617 \\
Col Aves &     228.054 &    2318.509 &     816.528 &     844.063 &       3.731 &   36537.497 &     224.422 &      16.760 \\
Col STDs &       9.209 &      77.348 &       0.044 &       0.047 &       0.070 &     202.296 &       9.210 &       0.275 \\

   aveStackAVE=   224.422
   rmsStackAVE=    10.279
   varStackAVE=   105.657

   aveStackRMS=    15.719
   rmsStackRMS=     3.613
   varStackRMS=    13.051

   aveStackVAR=   260.126
   rmsStackVAR=   115.565
   varStackVAR= 13355.253

ave = 224.422
rms =  19.125

FnF_BiasSubtracted_GFAv17.dat
GFAv17       Bkg/pix  PeakAboveBkg      Xpeak         Ypeak           FWHM       Total       BkgAve        BkgRMS
002 &     226.319 &    2421.338 &     816.576 &     844.113 &       3.639 &   36333.840 &     222.458 &      16.142 \\
003 &     220.627 &    2374.759 &     816.580 &     844.100 &       3.673 &   36307.734 &     217.039 &      16.697 \\
004 &     225.812 &    2383.223 &     816.567 &     844.119 &       3.682 &   36602.364 &     222.268 &      16.584 \\
005 &     228.907 &    2392.124 &     816.566 &     844.115 &       3.659 &   36285.915 &     225.303 &      16.808 \\
006 &     234.707 &    2369.898 &     816.543 &     844.079 &       3.684 &   36445.522 &     230.872 &      16.764 \\
007 &     231.458 &    2332.535 &     816.541 &     844.073 &       3.730 &   36770.214 &     228.122 &      16.961 \\
008 &     243.319 &    2219.559 &     816.494 &     844.023 &       3.829 &   36880.765 &     239.713 &      17.144 \\
009 &     239.735 &    2212.946 &     816.479 &     844.003 &       3.828 &   36746.992 &     236.029 &      16.757 \\
010 &     212.732 &    2248.784 &     816.471 &     844.004 &       3.780 &   36399.720 &     208.822 &      17.122 \\
011 &     216.920 &    2229.920 &     816.465 &     844.005 &       3.806 &   36601.908 &     213.593 &      16.617 \\
Col Aves &     228.054 &    2318.509 &     816.528 &     844.063 &       3.731 &   36537.497 &     224.422 &      16.760 \\
Col STDs &       9.209 &      77.348 &       0.044 &       0.047 &       0.070 &     202.296 &       9.210 &       0.275 \\
aveStackAVE=  224.422    rmsStackAVE=   10.279    aveStackRMS=   15.719    rmsStackRMS=    3.613


---------
FnF_NotBiasSubtracted_GFAv17.dat
GFAv17      Bkg/pix  PeakAboveBkg      Xpeak          Ypeak         FWHM       Total          BkgAve        BkgRMS
002 &    2156.695 &    2417.888 &     816.573 &     844.115 &       3.644 &   36381.552 &    2152.989 &      19.180 \\
003 &    2151.003 &    2371.363 &     816.578 &     844.102 &       3.678 &   36356.014 &    2147.570 &      19.578 \\
004 &    2156.186 &    2379.702 &     816.564 &     844.121 &       3.687 &   36653.954 &    2152.798 &      19.384 \\
005 &    2159.283 &    2388.762 &     816.563 &     844.117 &       3.664 &   36333.826 &    2155.834 &      19.882 \\
006 &    2165.082 &    2366.559 &     816.541 &     844.080 &       3.689 &   36494.547 &    2161.403 &      19.970 \\
007 &    2161.831 &    2328.996 &     816.538 &     844.075 &       3.735 &   36822.986 &    2158.653 &      20.563 \\
008 &    2173.692 &    2216.486 &     816.491 &     844.025 &       3.835 &   36934.081 &    2170.243 &      20.377 \\
009 &    2170.108 &    2209.893 &     816.476 &     844.005 &       3.834 &   36800.024 &    2166.560 &      19.670 \\
010 &    2143.107 &    2245.790 &     816.468 &     844.006 &       3.785 &   36449.641 &    2139.352 &      19.093 \\
011 &    2147.295 &    2226.954 &     816.462 &     844.007 &       3.811 &   36653.013 &    2144.124 &      19.085 \\
Col Aves &    2158.428 &    2315.239 &     816.526 &     844.065 &       3.736 &   36587.964 &    2154.953 &      19.678 \\
Col STDs &       9.208 &      77.161 &       0.044 &       0.047 &       0.070 &     204.314 &       9.210 &       0.494 \\
aveStackAVE= 2154.953    rmsStackAVE=   14.566    aveStackRMS=   15.719    rmsStackRMS=    3.613

   aveStackAVE=  2154.953
   rmsStackAVE=    14.566
   varStackAVE=   212.169

   aveStackRMS=    15.719
   rmsStackRMS=     3.613
   varStackRMS=    13.051

   aveStackVAR=   260.126
   rmsStackVAR=   115.565
   varStackVAR= 13355.253



"""



