# SIMv20.py
#  This makes noisy images and fits them.
#   26 Oct 2018 M.Lampton UCB SSL




import numpy as np
import matplotlib.pyplot as plt  
from scipy.optimize import curve_fit

PROGNAME = 'SIMv20.py'


GAIN      = 0.274     # ADU/electron from R.Casas study GFA05 quadrant 3
RMSADUBKG = 16.0       # ADU rms, overwritten below
OFFSET    = 200.0     # ADU      any constant goes here
PEAKADU   = 143.0    # overwritten below;  2270 ADU peak gives Rmag=14.00; 17mag=143 ADU peak.
XBAR      = 20.44
YBAR      = 19.21 
FWHM      = 3.74      # observed PSF, pixels

trueParms = np.array([OFFSET, PEAKADU, XBAR, YBAR, FWHM])


#-------THUMBNAIL SPECIFICS 40 x 40--------------------------------------


NPARMS = 5    # for fitting a 2D Gaussian
headerstr = '     Bkg/pix  PeakAboveBkg  Xpeak      Ypeak        FWHM      Total       1200Ave       1200RMS'




#-------define the adjustable image model for parameter fitting----------------------------

def getGridList(nside):
    # returns a list of tuples
    results = list()
    if nside <= 1:
        results.append((0.0, 0.0)) 
        return results
    center = 0.5*(nside-1)
    step = 2.0/nside
    nsq = nside**2
    for i in range(nsq):
        irow = i // nside
        icol = i % nside
        row = step*(irow - center)
        col = step*(icol - center)
        results.append((row,col))
    return results
    
def mySmoothFunc2D(parms, nrows, ncols):
    # MUST NOT IGNORE ANY PARMS OR THE MINIMIZER WILL FAIL
    # Smoooth Gaussian atop smooth backgrond, copied from GFAv17.py; here parms = fiveParms
    # Uses 25 samples per pixel, using devs[] from center {col,row}.
    # Produces a noiseless image, nrows x ncols, from given parameters.
    # parms[]: {background/pix;  peakSignal/pix; xpeak; ypeak; fwhmpix}
    denom = 2.*(parms[4]/2.35482)**2
    result = np.zeros((nrows,ncols))
    result.fill(parms[0])
    gridList = getGridList(NSUBS)
    nsubs = len(gridList)
    subvalues = np.zeros(nsubs)
    for row in range(nrows):
        for col in range(ncols):
            r2 = (col-parms[2])**2 + (row-parms[3])**2
            arg = -r2/denom
            if arg > -10.0:
                for isub in range(nsubs):
                    myTuple = gridList[isub]   # extract a tuple
                    drow = myTuple[0]          # extract a number
                    dcol = myTuple[1]          # extract the other number
                    r2 = (col + dcol - parms[2])**2 + (row + drow - parms[3])**2
                    arg = -r2/denom
                    subvalues[isub] = np.exp(arg)
                result[row,col] += parms[1]*np.mean(subvalues)  # NOTICE THE "+="
    return result


def createSimpleNoisyImage():
    # uses trueParms
    thumb2D = mySmoothFunc2D(trueParms, 40,40)
    thumb2D += np.random.normal(0., RMSADUBKG, (40,40))
    return thumb2D
    
def createNoisyImage():
    # uses trueParms
    thumb2D = mySmoothFunc2D(trueParms, 40,40)
    star2D = thumb2D - trueParms[0]   # remove constant bkg to get smooth star signal
    rootGain = np.sqrt(GAIN)
    rms2D = rootGain * np.sqrt(star2D)
    thumb2D += np.random.normal(0., RMSADUBKG, (40,40))  # needs explicit size
    thumb2D += np.random.normal(0., rms2D)  # automatically broadcasts size
    intThumb2D = thumb2D.astype(int)
    return intThumb2D
    
def mapThumbnail(rawADU):
    fig = plt.figure(figsize=(6,6))   # inches
    image = plt.imshow(rawADU)
    plt.colorbar(image, fraction=0.046, pad=0.04) # these "magically" work
    title = 'rawADUimage'
    plt.title(title)
    plt.savefig(title + '.png')
    plt.show()

def listThumbnail(rawADU):
    peak1D = rawADU.max()
    arg1D = rawADU.ravel().argmax()
    # arg1D = raw1D.argmax()
    sum = np.sum(rawADU)
    ave = np.mean(rawADU)
    med = np.median(rawADU)
    max = np.max(rawADU)
    min = np.min(rawADU)
    sumstr  = '   Sum:{:12.0f}'.format(sum)
    avestr  = '   Ave:{:8.3f}'.format(ave)
    medstr  = '   Med:{:7.1f}'.format(med)
    maxstr  = '   Max:{:5d}'.format(max)
    minstr  = '   Min:{:5d}'.format(min)
    peakstr = '   Peak:{:5d}'.format(peak1D)
    argstr  = '   Arg:{:5d}'.format(arg1D)
    sevenStr = sumstr + avestr + medstr + maxstr + minstr + peakstr + argstr
    print(sevenStr)




#---Optimization helpers: user should call this minimize2D() not curve_fit()-----
    
def minimize2D(goals2D, pinit1D, sigmas2D):
    # This is called by user to minimize.  
    #  It relies on wrapfunc1D() to always call myFunc2D().
    # Converts user input 2D arrays into 1D for calling curve_fit().
    # Also provides required args1D[] but we ignore them and use pixel {row,col}
    args1D = np.arange(1600) 
    goals1D = goals2D.reshape(1600)
    sigmas1D = sigmas2D.reshape(1600)
    popt, pcov = curve_fit(wrapfunc1D, args1D, goals1D, pinit1D, sigmas1D)
    return popt, pcov
    
def wrapfunc1D(args, *parms): 
    # Removing the asterisk below gives us our parameter array.
    # args[] is discarded since we use pixel {row,col} as arguments.  
    # global GNROWS, GNCOLS must be preset by minimize2D().
    global gncalls
    gncalls += 1
    model2D = mySmoothFunc2D(parms, 40, 40)
    model1D = np.reshape(model2D, 1600)
    return model1D

#--------LSQ fitter uses minimize2D()------------------------

def getBestFitParms(thumb2D):  # goals are noisy thumb2D
    global gncalls 
    # print('...Starting the fit')
    gncalls = 0  
    myPinit    = np.array([200, 200, 20, 20, 5.]) 
    mySigmas2D = 20*np.ones([40,40])       
    optim, covar = minimize2D(thumb2D, myPinit, mySigmas2D) 
    # print('ncalls = ', gncalls)
    totalFit = 1.1331 * optim[1] * optim[4]**2 
    result = np.zeros(6)
    for i in range(optim.size):
        result[i] = optim[i]
    result[5] = totalFit
    return result



#----three background zone locators-----------------

def build1200():   # full 1200 1D pixel bkg
    pixList = list()    
    for ipix in range(1600): 
        row, col = np.unravel_index(ipix, (40, 40)) 
        if row<10 or row>=30 or col<10 or col>=30:
            pixList.append(ipix)
    print('build1200() found N=', len(pixList))
    return pixList

def build700():       # outermost 700 1D pixels of the 1200
    # construct a list of 1D pixel addresses into 40x40 image
    pixList = list() 
    for ipix in range(1600):  
        row, col = np.unravel_index(ipix, (40, 40))
        if row<5 or row>=35 or col<5 or col>=35:
            pixList.append(ipix)
    print('build700() found N=', len(pixList))
    return pixList

def build500():    # innermost 500 1D pixels of the 1200
    # construct a list of 1D pixel addresses into 40x40 image
    pixList = list()    
    for ipix in range(1600): 
        row, col = np.unravel_index(ipix, (40, 40))
        if row>=5 and row<35 and col>=5 and col<35:
            if row<10 or row>=30 or col<10 or col>=30: 
                pixList.append(ipix)
    print('build500() found N=', len(pixList))
    return pixList




#--------------------------main program---------------------


print()
rawADU = np.array((40,40))
parms  = np.array(5)

"""#----lets do two PTCs: quiet and noisy------------

NSUBS = 4       # subdivide each pixel into 4x4
NITERS = 100    # do 100 frames  
NFLUXES = 40    # span 0 to 4000 ADU peak star signal
NNOISES = 2     # low noise = 6ADU, high noise = 16ADU
bkgRMSs = np.array([6.0, 16.0])

peakMeans = np.zeros((NFLUXES, NNOISES))
peakVars  = np.zeros((NFLUXES, NNOISES))

for inoise in range(NNOISES):
    RMSADUBKG = bkgRMSs[inoise]
    for iflux in range(NFLUXES):
        print('Starting iflux = ', iflux)
        truePeak = 100*iflux
        trueParms[1] = truePeak
        peaks = np.zeros(NITERS)
        for iter in range(NITERS):
            rawADU = createNoisyImage()
            raw1D = rawADU.ravel()
            peaks[iter] = raw1D[780]
        peakMeans[iflux,inoise] = np.mean(peaks)
        peakVars[iflux,inoise] =  np.var(peaks)
    
#--------plot these PTCs and their linear trends----------

fit0 = np.polyfit(peakMeans[:,0], peakVars[:,0], deg=1) # this gets the best fit coefs
slopeStr0 = 'for BkgRMS=6ADU: Slope, ADU/electron = {:5.3f}'.format(fit0[0])
fitFunc0 = np.poly1d(fit0)  # this is the callable polynomial

fit1 = np.polyfit(peakMeans[:,1], peakVars[:,1], deg=1) # this gets the best fit coefs
slopeStr1 = 'for BkgRMS=16ADU: Slope,ADU/electron = {:5.3f}'.format(fit1[0])
fitFunc1 = np.poly1d(fit1)  # this is the callable polynomial

fig, ax = plt.subplots()
plt.plot(peakMeans[:,0], peakVars[:,0], 'bo', peakMeans[:,0], fitFunc0(peakMeans[:,0]), '-b')
plt.plot(peakMeans[:,1], peakVars[:,1], 'ro', peakMeans[:,1], fitFunc1(peakMeans[:,1]), '-r')
ax.tick_params(direction='in')
ax.set_xlabel('Mean ADU, 100 images, for pixel 780')
ax.set_ylabel('Variance, ADU, 100 images, for pixel 780')
ax.text(0.05, 0.85, slopeStr0, transform=ax.transAxes, fontsize=10, color='blue')
ax.text(0.05, 0.90, slopeStr1, transform=ax.transAxes, fontsize=10, color='red')
ax.text(0.8, 0.05, PROGNAME, transform=ax.transAxes, fontsize=8)
plt.title('PTC for Simulated Data')

plt.savefig('PTC-SIMv20.png') 
plt.show()           
"""

#------perform NITERS successive fits----------

NSUBS = 4       # subdivide each pixel into 4x4
NITERS = 100

catcher = np.zeros((NITERS, 6))  # catch the results for plotting or analysis

header = '      Back        Peak         Xbar        Ybar        FWHM   RMS & TOTAL'

initParms = np.zeros(6)
for i in range(5):
    initParms[i] = trueParms[i]
initParms[5] = RMSADUBKG    # not really a control parm: it is a noise input

initStr = ''
for i in range(initParms.size):
    initStr += '{:12.5f}'.format(initParms[i]) 
print(header)
print(initStr)
    
rmsStr = str(RMSADUBKG)
outfilename = 'SIMv20-LSQ-' + rmsStr + '.dat'    
with open(outfilename, 'w') as outfile:
    outfile.writelines(header + '\n')   # yes plural even for one line
    outfile.writelines(initStr + '\n')  # yes plural even for one line
    for i in range(NITERS):
        rawADU = createNoisyImage()
        # listThumbnail(rawADU)
        # mapThumbnail(rawADU)
        sixParms = getBestFitParms(rawADU)
        catcher[i] = sixParms
        sixStr = ''
        for i in range(sixParms.size):
            sixStr += '{:12.4f}'.format(sixParms[i]) 
        outfile.writelines(sixStr + '\n')  # yes plural even for one line
        print(sixStr)
        
shapeTuple = catcher.shape
nrows = shapeTuple[0]
ncols = shapeTuple[1]
print('nrows, ncols = ', nrows, ncols)

for row in range(nrows):
    for col in range(ncols):
        print('{:12.4f}'.format(catcher[row,col]), end='')
    print()

aves = np.zeros(ncols)
rmss = np.zeros(ncols)
for col in range(ncols):
    aves[col] = np.mean(catcher[:,col])
    rmss[col] = np.std(catcher[:,col])

print()
print('Aves:')
for col in range(ncols):
    print('{:12.4f}'.format(aves[col]), end='')
print()
print('RMSs:')
for col in range(ncols):
    print('{:12.4f}'.format(rmss[col]), end='')
print()



