# SIMv23.py
#  This makes noisy images and fits them.
#  100 iters not 1000
#  seeing = 0.8 arcsec (FWHM=3.75pix) also seeing = 1.6 arcsec (FWHM=7.5pixels)
#  BkgNoise = 6ADU = 22 erms and 16ADU=58 erms
#  Rmag = 13, 13.5, 14., 14.5, 15, 15.5, 16, 16.5, 17
#   27 Oct 2018 M.Lampton UCB SSL




import numpy as np
import matplotlib.pyplot as plt  
from scipy.optimize import curve_fit

PROGNAME = 'SIMv23.py'


GAIN      = 0.274     # ADU/electron from R.Casas study GFA05 quadrant 3
RMSADUBKG = 16.0       # ADU rms, overwritten below
OFFSET    = 200.0     # ADU      any constant goes here
PEAKADU   = 143.0    # overwritten below;  2270 ADU peak gives Rmag=14.00; 17mag=143 ADU peak.
XBAR      = 20.44
YBAR      = 19.21 
FWHM      = 3.74      # observed PSF, pixels

trueParms = np.array([OFFSET, PEAKADU, XBAR, YBAR, FWHM])


#-------THUMBNAIL SPECIFICS 40 x 40--------------------------------------


NPARMS = 5    # for fitting a 2D Gaussian
headerstr = '     Bkg/pix  PeakAboveBkg  Xpeak      Ypeak        FWHM      Total       1200Ave       1200RMS'




#-------define the adjustable image model for parameter fitting----------------------------

def getGridList(nside):
    # returns a list of tuples
    results = list()
    if nside <= 1:
        results.append((0.0, 0.0)) 
        return results
    center = 0.5*(nside-1)
    step = 2.0/nside
    nsq = nside**2
    for i in range(nsq):
        irow = i // nside
        icol = i % nside
        row = step*(irow - center)
        col = step*(icol - center)
        results.append((row,col))
    return results
    
def mySmoothFunc2D(parms, nrows, ncols):
    # MUST NOT IGNORE ANY PARMS OR THE MINIMIZER WILL FAIL
    # Smoooth Gaussian atop smooth backgrond, copied from GFAv17.py; here parms = fiveParms
    # Uses 25 samples per pixel, using devs[] from center {col,row}.
    # Produces a noiseless image, nrows x ncols, from given parameters.
    # parms[]: {background/pix;  peakSignal/pix; xpeak; ypeak; fwhmpix}
    denom = 2.*(parms[4]/2.35482)**2
    result = np.zeros((nrows,ncols))
    result.fill(parms[0])
    gridList = getGridList(NSUBS)
    nsubs = len(gridList)
    subvalues = np.zeros(nsubs)
    for row in range(nrows):
        for col in range(ncols):
            r2 = (col-parms[2])**2 + (row-parms[3])**2
            arg = -r2/denom
            if arg > -10.0:
                for isub in range(nsubs):
                    myTuple = gridList[isub]   # extract a tuple
                    drow = myTuple[0]          # extract a number
                    dcol = myTuple[1]          # extract the other number
                    r2 = (col + dcol - parms[2])**2 + (row + drow - parms[3])**2
                    arg = -r2/denom
                    subvalues[isub] = np.exp(arg)
                result[row,col] += parms[1]*np.mean(subvalues)  # NOTICE THE "+="
    return result


def createSimpleNoisyImage():
    # uses trueParms
    thumb2D = mySmoothFunc2D(trueParms, 40,40)
    thumb2D += np.random.normal(0., RMSADUBKG, (40,40))
    return thumb2D
    
def createNoisyImage():
    # uses trueParms
    thumb2D = mySmoothFunc2D(trueParms, 40,40)
    star2D = thumb2D - trueParms[0]   # remove constant bkg to get smooth star signal
    rootGain = np.sqrt(GAIN)
    rms2D = rootGain * np.sqrt(star2D)
    thumb2D += np.random.normal(0., RMSADUBKG, (40,40))  # needs explicit size
    thumb2D += np.random.normal(0., rms2D)  # automatically broadcasts size
    intThumb2D = thumb2D.astype(int)
    return intThumb2D
    
def mapThumbnail(rawADU):
    fig = plt.figure(figsize=(6,6))   # inches
    image = plt.imshow(rawADU)
    plt.colorbar(image, fraction=0.046, pad=0.04) # these "magically" work
    title = 'rawADUimage'
    plt.title(title)
    plt.savefig(title + '.png')
    plt.show()

def listThumbnail(rawADU):
    peak1D = rawADU.max()
    arg1D = rawADU.ravel().argmax()
    # arg1D = raw1D.argmax()
    sum = np.sum(rawADU)
    ave = np.mean(rawADU)
    med = np.median(rawADU)
    max = np.max(rawADU)
    min = np.min(rawADU)
    sumstr  = '   Sum:{:12.0f}'.format(sum)
    avestr  = '   Ave:{:8.3f}'.format(ave)
    medstr  = '   Med:{:7.1f}'.format(med)
    maxstr  = '   Max:{:5d}'.format(max)
    minstr  = '   Min:{:5d}'.format(min)
    peakstr = '   Peak:{:5d}'.format(peak1D)
    argstr  = '   Arg:{:5d}'.format(arg1D)
    sevenStr = sumstr + avestr + medstr + maxstr + minstr + peakstr + argstr
    print(sevenStr)




#---Optimization helpers: user should call this minimize2D() not curve_fit()-----
    
def minimize2D(goals2D, pinit1D, sigmas2D):
    # This is called by user to minimize.  
    #  It relies on wrapfunc1D() to always call myFunc2D().
    # Converts user input 2D arrays into 1D for calling curve_fit().
    # Also provides required args1D[] but we ignore them and use pixel {row,col}
    args1D = np.arange(1600) 
    goals1D = goals2D.reshape(1600)
    sigmas1D = sigmas2D.reshape(1600)
    popt, pcov = curve_fit(wrapfunc1D, args1D, goals1D, pinit1D, sigmas1D)
    return popt, pcov
    
def wrapfunc1D(args, *parms): 
    # Removing the asterisk below gives us our parameter array.
    # args[] is discarded since we use pixel {row,col} as arguments.  
    # global GNROWS, GNCOLS must be preset by minimize2D().
    global gncalls
    gncalls += 1
    model2D = mySmoothFunc2D(parms, 40, 40)
    model1D = np.reshape(model2D, 1600)
    return model1D

#--------LSQ fitter uses minimize2D()------------------------

def getBestFitParms(thumb2D):  # goals are noisy thumb2D
    global gncalls 
    # print('...Starting the fit')
    gncalls = 0  
    myPinit    = np.array([200, 200, 20, 20, 5.]) 
    mySigmas2D = 20*np.ones([40,40])       
    optim, covar = minimize2D(thumb2D, myPinit, mySigmas2D) 
    # print('ncalls = ', gncalls)
    totalFit = 1.1331 * optim[1] * optim[4]**2 
    result = np.zeros(6)
    for i in range(optim.size):
        result[i] = optim[i]
    result[5] = totalFit
    return result



#----three background zone locators-----------------

def build1200():   # full 1200 1D pixel bkg
    pixList = list()    
    for ipix in range(1600): 
        row, col = np.unravel_index(ipix, (40, 40)) 
        if row<10 or row>=30 or col<10 or col>=30:
            pixList.append(ipix)
    print('build1200() found N=', len(pixList))
    return pixList

def build700():       # outermost 700 1D pixels of the 1200
    # construct a list of 1D pixel addresses into 40x40 image
    pixList = list() 
    for ipix in range(1600):  
        row, col = np.unravel_index(ipix, (40, 40))
        if row<5 or row>=35 or col<5 or col>=35:
            pixList.append(ipix)
    print('build700() found N=', len(pixList))
    return pixList

def build500():    # innermost 500 1D pixels of the 1200
    # construct a list of 1D pixel addresses into 40x40 image
    pixList = list()    
    for ipix in range(1600): 
        row, col = np.unravel_index(ipix, (40, 40))
        if row>=5 and row<35 and col>=5 and col<35:
            if row<10 or row>=30 or col<10 or col>=30: 
                pixList.append(ipix)
    print('build500() found N=', len(pixList))
    return pixList




#--------------------------main program---------------------


print()
rawADU = np.array((40,40))
parms  = np.array(5)


#------perform NITERS successive fits----------

NSUBS = 4       # subdivide each pixel into 4x4
NITERS = 100

Rmags   = np.array([13.0,   13.5, 14.0,   14.5,  15.0, 15.5, 16.0, 16.5, 17.0])
peaks  = np.array([[5706., 3600., 2270., 1433., 904., 571., 360., 227., 143.],  # for 0.8 arcsec seeing
                   [1426., 900.,  567.,  358., 226., 143.,  90.,  57.,  36.]])  # for 1.6 arcsec seeing
seeings = np.array([0.8, 1.6])
noises  = np.array([6.0, 16.0])


nseeings = seeings.size
nmags    = Rmags.size
nnoises  = noises.size

catcher = np.zeros((NITERS, 6))  # catch the results for one iteration

header = ' BkgRMS  seeing  Rmag   Peak '
header += '      Back        Peak         Xbar        Ybar        FWHM     TOTAL  '
header += '      BackRMS     PeakRMS     XbarRMS     YbarRMS    FWHM_RMS   TOTAL_RMS'


print(header)
outfilename = 'SIMv23-multi.dat'    
with open(outfilename, 'w') as outfile:
    outfile.writelines(header +'\n')  # yes plural even for one line
    for iseeing in range(nseeings):
        see = seeings[iseeing]            # arcsec
        seeStr = '{:7.2f}'.format(see)    # arcsec
        trueParms[4] = see/0.214          # FWHM
        for inoise in range(nnoises):
            RMSADUBKG = noises[inoise]    # global
            bkgStr = '{:7.1f}'.format(RMSADUBKG)
            for imag in range(nmags):
                trueParms[1] =  peaks[iseeing, imag]  
                RmagStr = '{:7.1f}'.format(Rmags[imag])
                peakStr = '{:8.1f}'.format(trueParms[1])
                for i in range(NITERS):
                    rawADU = createNoisyImage()
                    sixParms = getBestFitParms(rawADU)
                    catcher[i] = sixParms        
                shapeTuple = catcher.shape
                nrows = shapeTuple[0]
                ncols = shapeTuple[1]
                aves = np.zeros(ncols)
                rmss = np.zeros(ncols)
                aveStr = ''
                rmsStr = ''            
                for col in range(ncols):
                    aves[col] = np.mean(catcher[:,col])
                    rmss[col] = np.std(catcher[:,col])
                    aveStr += '{:12.4f}'.format(aves[col])
                    rmsStr += '{:12.4f}'.format(rmss[col])
                allStr = bkgStr + seeStr + RmagStr + peakStr + aveStr + rmsStr
                outfile.writelines(allStr + '\n')  # yes plural even for one line 
                print(allStr)




